﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Emgu.CV;
using Emgu.CV.Structure;
using bbiwarg.Images;
using bbiwarg.Utility;

namespace bbiwarg.Detectors.Fingers
{
    class Finger
    {
        private Vector2D tipPoint;
        private Vector2D handPoint;
        private Vector2D direction;
        private float length;

        public Finger(Vector2D p1, Vector2D p2) {
            assignTipAndHandPoint(p1, p2);
        }

        public float getLength() {
            return length;
        }

        public Vector2D getTipPoint() {
            return tipPoint;
        }

        public Vector2D getHandPoint() {
            return handPoint;
        }

        public Vector2D getDirection() {
            return direction;
        }

        public float getSimilarity(Finger compareFinger)
        {
            //thresholds
            float thresholdMaxAngle = (float)(30 * Math.PI / 180); // 30°
            float thresholdMaxParallelDistance = 40;
            float thresholdMaxVerticalDistance = 40;

            //check angle
            float angle = direction.getMinAngleBetween(compareFinger.direction);
            float angleSimilarity = Math.Max(1 - angle / thresholdMaxAngle, 0);

            //check parallel distance
            float minParallelDistance = getMinParallelDistance(compareFinger);
            float parallelDistanceSimilarity = Math.Max(1 - minParallelDistance / thresholdMaxParallelDistance, 0);

            //check vertical distance
            float minVerticalDistance = getMinVerticalDistance(compareFinger);
            float verticalDistanceSimilarity = Math.Max(1 - minVerticalDistance / thresholdMaxVerticalDistance, 0);

            return (angleSimilarity + parallelDistanceSimilarity + verticalDistanceSimilarity) / 3;
        }

        public bool isMergeable(Finger mergeFinger)
        {
            //thresholds
            float thresholdMaxAngle = (float) (15 * Math.PI / 180); // 15°
            float thresholdMaxParallelDistance = 10;
            float thresholdMaxVerticalDistance = 20;

            //check angle
            float angle = direction.getMinAngleBetween(mergeFinger.direction);
            if (angle > thresholdMaxAngle) return false;

            //check parallel distance
            float minParallelDistance = getMinParallelDistance(mergeFinger);
            if (minParallelDistance > thresholdMaxParallelDistance) return false;

            //check vertical distance
            float minVerticalDistance = getMinVerticalDistance(mergeFinger);
            if (minVerticalDistance > thresholdMaxVerticalDistance) return false;

            return true;
        }

        public void mergeFingers(Finger mergeFinger) {
            Vector2D a1 = tipPoint;
            Vector2D a2 = handPoint;
            Vector2D b1 = mergeFinger.getTipPoint();
            Vector2D b2 = mergeFinger.getHandPoint();

            //find two points that are furthest apart
            float distanceA1A2 = length;
            float distanceA1B1 = a1.getDistanceTo(b1);
            float distanceA1B2 = a1.getDistanceTo(b2);
            float distanceA2B1 = a2.getDistanceTo(b1);
            float distanceA2B2 = a2.getDistanceTo(b2);
            float distanceB1B2 = mergeFinger.getLength();
            float maxDistance = Math.Max(Math.Max(distanceA1A2, distanceA1B1), Math.Max(Math.Max(distanceA1B2, distanceA2B1), Math.Max(distanceA2B2, distanceB1B2)));

            if(maxDistance == distanceA1A2) assignTipAndHandPoint(a1,a2);
            else if(maxDistance == distanceA1B1) assignTipAndHandPoint(a1, b1);
            else if(maxDistance == distanceA1B2) assignTipAndHandPoint(a1, b2);
            else if(maxDistance == distanceA2B1) assignTipAndHandPoint(a2, b1);
            else if(maxDistance == distanceA2B2) assignTipAndHandPoint(a2, b2);
            else if(maxDistance == distanceB1B2) assignTipAndHandPoint(b1, b2);
        }

        private void assignTipAndHandPoint(Vector2D p1, Vector2D p2)
        {
            if (p1.y < p2.y)
            {
                this.tipPoint = p1;
                this.handPoint = p2;
            }
            else
            {
                this.tipPoint = p2;
                this.handPoint = p1;
            }

            direction = handPoint - tipPoint;
            length = direction.getLength();
        }

        private float getMinParallelDistance(Finger finger)
        {
            Vector2D a1 = tipPoint;
            Vector2D a2 = handPoint;
            Vector2D b1 = finger.getTipPoint();
            Vector2D b2 = finger.getHandPoint();

            //project on finger line
            Vector2D p1 = Vector2D.projectToLine(b1, direction, a1);
            Vector2D p2 = Vector2D.projectToLine(b2, direction, a1);

            float distanceP1B1 = p1.getDistanceTo(b1);
            float distanceP2B2 = p2.getDistanceTo(b2);
            return Math.Min(distanceP1B1, distanceP2B2);
        }

        private float getMinVerticalDistance(Finger finger)
        {
            Vector2D a1 = tipPoint;
            Vector2D a2 = handPoint;
            Vector2D b1 = finger.getTipPoint();
            Vector2D b2 = finger.getHandPoint();

            //project on finger line
            Vector2D p1 = Vector2D.projectToLine(b1, direction, a1);
            Vector2D p2 = Vector2D.projectToLine(b2, direction, a1);

            float distanceP1A1 = p1.getDistanceTo(a1);
            float distanceP1A2 = p1.getDistanceTo(a2);
            float distanceP2A1 = p2.getDistanceTo(a1);
            float distanceP2A2 = p2.getDistanceTo(a2);
            return Math.Min(Math.Min(distanceP1A1, distanceP1A2), Math.Min(distanceP2A1, distanceP2A2));
        }
    }
}
